// InputMode.cpp : Defines the entry point for the console application.
//


#include "stdafx.h"

//----------------------------------------------------------
// Global Variables 
//----------------------------------------------------------

FT_STATUS ftStatus;
FT_HANDLE ftHandle;
Receiver_States RxStates;
UCHAR RxFrameBuf[256];
int idxRxFrame;

//----------------------------------------------------------
// Functions prototypes (Public)
//----------------------------------------------------------

DWORD GetDeviceList(DevList * list);
DWORD PrintDeviceList(DevList * list);
DWORD ChooseDeviceFromList(DevList * list);
void FreeAllocatedMemory(DevList * list);
void DisconnectDevice();
void Set_GPIO0(FT_HANDLE device_handle);
void Reset_GPIO0(FT_HANDLE device_handle);
void Set_GPIO1(FT_HANDLE device_handle);
void Reset_GPIO1(FT_HANDLE device_handle);
void Send_ResetRequest(FT_HANDLE device_handle);
void Send_EEPROMDumpRequest(FT_HANDLE device_handle);
void PrintHelp();
void ProcessFrame();
void ProcessWiegandFrame();
void ProcessTamperFrame();
void ProcessRxData(unsigned char * RxChars, unsigned int BytesRx);

//----------------------------------------------------------
// Functions prototypes (Private)
//----------------------------------------------------------

//void Send_Command(FT_HANDLE device_handle, UCHAR cmd_ID, UCHAR cmd_data_size, UCHAR * cmd_data);

//----------------------------------------------------------
// Main code
//----------------------------------------------------------
int _tmain(int argc, _TCHAR* argv[])
{
	DWORD numAvailableDevs=0;
	int rValue;
	char ch;
	UCHAR RxBuffer[256];
	DWORD RxBytes,BytesReceived;
	static DevList AvailableDevicesList;

	printf("ETConcept Engenharia de sistemas LDA\n");
	printf("W2USB Wiegand Converters Application\n");
	printf("Example 1 - Input mode\n");
	printf("\n\n List of avaiable devices: \n\n");

	numAvailableDevs=GetDeviceList(&AvailableDevicesList);

	PrintDeviceList(&AvailableDevicesList);

	rValue=ChooseDeviceFromList(&AvailableDevicesList);

	PrintHelp();

	while(rValue != -1){
		if (_kbhit()){
			ch = _getch();
			switch(ch){
			case 'O':
				printf("Set GPIO0");
				Set_GPIO0(ftHandle);
				break;
			case 'o':
				printf("Reset GPIO0");
				Reset_GPIO0(ftHandle);
				break;
			case 'P':
				printf("Set GPIO1");
				Set_GPIO1(ftHandle);
				break;
			case 'p':
				printf("Reset GPIO1");
				Reset_GPIO1(ftHandle);
				break;
			case 'h':
			case 'H':
				PrintHelp();
				break;
			case 'r':
			case 'R':
				printf("Reset request ");
				Send_ResetRequest(ftHandle);
				break;
			case 'd':
			case 'D':
				printf("EEPROM Dump Request  ");
				Send_EEPROMDumpRequest(ftHandle);
				break;
			case 'x':
			case 'X':
				printf("Exit \n\n");
				rValue=-1;
				break;
			};
		}

		// Check if new data wassend by the converter

        // Do your seconds based tasks here.
		FT_GetQueueStatus(ftHandle,&RxBytes); 
		if (RxBytes>0){
			ftStatus = FT_Read(ftHandle,RxBuffer,RxBytes,&BytesReceived);
			ProcessRxData(RxBuffer,BytesReceived);
			ProcessFrame();
		}
	}

	FreeAllocatedMemory(&AvailableDevicesList);

	DisconnectDevice();
	return 0;
}

DWORD GetDeviceList(DevList * list){

	FT_STATUS ftStatus;
	FT_HANDLE ftHandleTemp;
	DWORD numDevs, auxnumDevs;
	DWORD Flags;
	DWORD ID;
	DWORD Type;
	DWORD LocId;
	UCHAR SerialNumber[16];
	UCHAR Description[64]; // create the device information list

	ftStatus = FT_CreateDeviceInfoList(&numDevs);
	auxnumDevs=0;
	while (auxnumDevs < numDevs && auxnumDevs<MAXIMUM_NUM_DEVICES ) { 
		ftStatus = FT_GetDeviceInfoDetail(auxnumDevs, &Flags, &Type, &ID, &LocId, SerialNumber, Description, &ftHandleTemp);
		if (ftStatus == FT_OK) { 
			list->SN_List[auxnumDevs]=new UCHAR[sizeof(SerialNumber)+1];
			strcpy_s((char *)list->SN_List[auxnumDevs],sizeof(SerialNumber),(char *)SerialNumber);
		}
		auxnumDevs++;
	}
	list->numDevs=numDevs;
	return numDevs;
}

DWORD PrintDeviceList(DevList * list){
	DWORD numDevs=0;

	while (numDevs < list->numDevs && list->SN_List[numDevs]!=NULL ) { 
		printf("%d) SN=%s \n",numDevs+1,list->SN_List[numDevs]);
		numDevs++;
	}
	return numDevs;		// Return thenumber of devices printed
}

DWORD ChooseDeviceFromList(DevList * list){

	int devIdx=-1;

	while(devIdx<1 || devIdx>(signed int)list->numDevs){
		printf("\n Choose from (1-%d) :",list->numDevs);
		scanf_s("%d",&devIdx);
	}

	// Connect to the Device using the SerialNumber
	ftStatus=FT_OpenEx(list->SN_List[devIdx-1],FT_OPEN_BY_SERIAL_NUMBER,&ftHandle);
	if (ftStatus != FT_OK){
		printf("Erro! Error code %d (USB Device %s isn't plug to USB port)\n\r",ftStatus,list->SN_List[devIdx-1]);
		return NULL;
	}

	// *************************************
	// Set the Device communication settings
	// *************************************

	//	Set BaudRate to 9600  (Baudrate should be maintain at 9600!)
	ftStatus = FT_SetBaudRate(ftHandle, 9600); // Set baud rate to 9600
	if (ftStatus != FT_OK){
		printf("Erro! Error code %d (Couldn't set the BaudRate)\n\r",ftStatus);
		return ftStatus;
	}
	//	Set Databits to 8bits, StopBits to 1bit and Parity to None
	ftStatus = FT_SetDataCharacteristics(ftHandle, FT_BITS_8, FT_STOP_BITS_1, FT_PARITY_NONE);
	if (ftStatus != FT_OK){
		printf("Erro! Error code %d (Couldn't set Data bits or Stop bits or Parity bit)\n\r",ftStatus);
		return ftStatus;
	}
	printf("\n Connected to device SN=%s\n",list->SN_List[devIdx-1]);
	return devIdx;
}

void FreeAllocatedMemory(DevList * list){
	for (int nums=list->numDevs; nums>0; nums--)
		delete [] list->SN_List[nums];
}


void DisconnectDevice(){

	if(ftHandle)							// If connected ato any device
		ftStatus=FT_Close(ftHandle);		// disconect
	if (ftStatus != FT_OK){
		
	}
}

void Send_Command(FT_HANDLE device_handle, UCHAR cmd_ID, UCHAR cmd_data_size, UCHAR * cmd_data){
	char TmpRxBuffer[21];
	DWORD BytesWritten;
	TmpRxBuffer[0]=0x55;
	TmpRxBuffer[1]=0x55;
	TmpRxBuffer[2]=cmd_ID;
	strcpy_s(TmpRxBuffer+3,cmd_data_size+1 ,(char *)cmd_data);
	TmpRxBuffer[cmd_data_size+4]=0x00;
	TmpRxBuffer[cmd_data_size+4]=0x0D;

	ftStatus = FT_Write(device_handle, TmpRxBuffer, cmd_data_size+5, &BytesWritten); 
	if (ftStatus == FT_OK) { // FT_Write OK 
		printf("  Ok\n");
	} else { 
	// FT_Write Failed 
		printf("  Nok\n");
	}
}

void Set_GPIO0(FT_HANDLE device_handle){
	UCHAR cmd_data[2];

	cmd_data[0]=0xFF;
	cmd_data[1]=0x00;

	Send_Command(device_handle, 0x03, 1, cmd_data);
}

void Reset_GPIO0(FT_HANDLE device_handle){
	UCHAR cmd_data[2];

	cmd_data[0]=0x00;
	cmd_data[1]=0x00;

	Send_Command(device_handle, 0x03, 1, cmd_data);
}

void Set_GPIO1(FT_HANDLE device_handle){
	UCHAR cmd_data[2];

	cmd_data[0]=0xFF;
	cmd_data[1]=0x00;

	Send_Command(device_handle, 0x04, 1, cmd_data);
}

void Reset_GPIO1(FT_HANDLE device_handle){
	UCHAR cmd_data[2];

	cmd_data[0]=0x00;
	cmd_data[1]=0x00;

	Send_Command(device_handle, 0x04, 1, cmd_data);
}

void Send_ResetRequest(FT_HANDLE device_handle){
	UCHAR cmd_data[2];

	cmd_data[0]=0x00;
	cmd_data[1]=0x00;

	Send_Command(device_handle, 0x0A, 1, cmd_data);
}

void Send_EEPROMDumpRequest(FT_HANDLE device_handle){
	UCHAR cmd_data[2];
	unsigned char RxBuffer[256];
	DWORD RxBytes,BytesReceived;

	// Clear Rx buffer
	FT_SetTimeouts(device_handle,1000,0);		// Set Read timeout to 1 second
	FT_GetQueueStatus(device_handle,&RxBytes); 
	if (RxBytes>0)
		ftStatus = FT_Read(device_handle,RxBuffer,RxBytes,&BytesReceived); 


	cmd_data[0]=0x00;
	cmd_data[1]=0x00;

	Send_Command(device_handle, 0x0B, 1, cmd_data);

	::Sleep(100);
	// Wait for the response
	FT_SetTimeouts(device_handle,1000,0);		// Set Read timeout to 1 second
	FT_GetQueueStatus(device_handle,&RxBytes); 
	if (RxBytes>0){
		ftStatus = FT_Read(ftHandle,RxBuffer,RxBytes,&BytesReceived); 
		if (ftStatus == FT_OK) { // FT_Read OK 
			// TODO: Do not assumes that received data is ok
			printf("\n");
			for (unsigned int i=0; i < RxBytes ; i++){
				printf("%02X ",(unsigned char)RxBuffer[i]);
				if ( ((i+1) % 13)== 0 ){
					printf("\n");
				}
			}
			printf("\n");
		} else { 
			// FT_Read Failed 
			printf("Error reading EEPROM \n");
		} 
	}

}

void PrintHelp(){
	printf("\n\n");
	printf ("Press 'O' to set GPIO0 to high level\n");
	printf ("Press 'o' to set GPIO0 to low level\n");
	printf ("Press 'P' to set GPIO1 to high level\n");
	printf ("Press 'p' to set GPIO1 to low level\n");
	printf ("Press 'r' or 'R' to send a Reset device\n");
	printf ("Press 'd' or 'D' to read the device EEPROM\n");
	printf ("Press 'x' or 'X' to exit\n");
	printf("\n\n");
}

void ProcessRxData(unsigned char * RxChars, unsigned int BytesRx){
	unsigned int i;	

	i=0;
	while (i<BytesRx){
		switch(RxStates){
			case WAIT_SYNC:
				// Verify if any of the received Chars are 0x55
				while(RxChars[i]!=(unsigned char)0x55 && i<BytesRx) i++;
				// If exist a 0x55 char Check if char next to it is also a 0x55
				if (RxChars[i]==(unsigned char)0x55 && BytesRx>i){
					// First Sync char detected
					// Copy to the RxFrameBuf
					RxFrameBuf[idxRxFrame++]=RxChars[i];
					// Check if the adjacent char is also Sync char
					if((i+1)<BytesRx){	// Sync char isn't the last of this block
						// Check if the adjacent char is also Sync char
						// If true, Preliminary Start of the frame detected change to WAIT_END state
						i++;
						if (RxChars[i]==(unsigned char)0x55){
							RxFrameBuf[idxRxFrame++]=RxChars[i++];
							RxStates=WAIT_END;
						}
					}else{
						// First Sync char detected in the end of this block wait for the second Sync
						RxStates=WAIT_SYNC2;
					}
				}else{
					// else maintain the current state
				}
				break;
			case WAIT_SYNC2:
				if (RxChars[i]==(unsigned char)0x55){
					RxFrameBuf[idxRxFrame++]=RxChars[i++];
					RxStates=WAIT_END;
				}else{
					idxRxFrame=0;			// False Alarm clear RxFrameBuf
					RxStates=WAIT_SYNC;
				}
				i++;
				break;
			case WAIT_END:
				// Check if any of the received chars is 0x0D
				while(RxChars[i]!=(unsigned char)0x0D && i<BytesRx && idxRxFrame<20){
					RxFrameBuf[idxRxFrame++]=RxChars[i++];	// Store Char
				}
				if (idxRxFrame>=20){
					// End of frame wasnt detected
					idxRxFrame=0;			// False Alarm clear RxFrameBuf
					// Return to Wait sync
					RxStates=WAIT_SYNC;
				}else if (i>=BytesRx){		// End of the block
					// Maintain Rx state
				}else{
					// End of frame detected
					RxFrameBuf[idxRxFrame++]=RxChars[i++];
					// check frame coerence Dimension x distance between 
					if ( ( RxFrameBuf[2] == (unsigned char)0x01 && idxRxFrame==20) ||
						 ( RxFrameBuf[2] == (unsigned char)0x02 && idxRxFrame==6) ||
						 ( RxFrameBuf[2] == (unsigned char)0x03 && idxRxFrame==6) ||
						 ( RxFrameBuf[2] == (unsigned char)0x04 && idxRxFrame==6) ||
						 ( RxFrameBuf[2] == (unsigned char)0x09 && idxRxFrame==7) ||
						 ( RxFrameBuf[2] == (unsigned char)0x0A && idxRxFrame==6) ||
						 ( RxFrameBuf[2] == (unsigned char)0x0B && idxRxFrame==6)
					   ){
						// Sync chars and end frame char and dim ok
						// Change to PROCESSING_DATA
						RxStates=PROCESSING_DATA;
					}else{
						// Frame isnt over. Maintain same state

					}
				}
			break;
			case PROCESSING_DATA:
				ProcessFrame();
			break;
		};
	}
}

void ProcessFrame(){
	if (RxStates==PROCESSING_DATA){
		switch(RxFrameBuf[2]){
			case 0x01:	// Wiegand Frame Received
				ProcessWiegandFrame();
				break;
			case 0x02:  // Change notification of the Tamper
				ProcessTamperFrame();
				break;

		// (Only for Output mode)
		//case 0x03:  // Change notification of the GPIO0
		//	ProcessGPIO0Frame();		
		//	break;
		//case 0x04:  // Change notification of the GPIO1
		//	ProcessGPIO1Frame();
		//	break;
		//case 0x09:	// Change EEPROM data
		//	// do nothing
		//	break;

			default:
				break;
		};
		idxRxFrame=0;
		RxStates=WAIT_SYNC;
	}
}


void ProcessWiegandFrame(){

	int wgn_dim;
	unsigned int wgn_facility_code;
	unsigned int wgn_badge_code;
	unsigned char wgn_parityE,wgn_parityO;

	unsigned long wgn_data, wgn_data_aux;
	void * ptr_wgn_ldata= &wgn_data;
	unsigned char * ptr_wgn_data=(unsigned char *)ptr_wgn_ldata;

	wgn_dim=(int)ceill((double)RxFrameBuf[3]/8.0);	// Wiegand Data bytes
	ptr_wgn_data+=(wgn_dim-1);

	for (int i=0; i < wgn_dim ; i++){
		*ptr_wgn_data= (unsigned char)RxFrameBuf[4+i];
		ptr_wgn_data--;
	}

	printf("Wiegand Frame Hex format:\n");
	for (int i=0; i < 20 ; i++){
		printf("%02X ", (unsigned char)RxFrameBuf[i]);
	}
	printf("\n");

	switch((unsigned char)RxFrameBuf[3]){
	case 8:			// 8 Bit Wiegand
		break;
	case 26:		// 26 Bit Wiegand
		printf("26 Bit Wiegand, ");
		wgn_facility_code= (wgn_data >> (17+6)) & 0x00FF;
		wgn_badge_code= (wgn_data >> (1+6)) & 0x7FFF;

		printf("Facility Code=%d, Badge ID=%d, ",wgn_facility_code,wgn_badge_code);

		// Check parity bits
		// Parity Even
		wgn_parityE=0;
		wgn_data_aux=wgn_data>>(13+6);
		for (int i=0; i<13; i++){
			wgn_parityE^=(wgn_data_aux & 0x00000001);
			wgn_data_aux=wgn_data_aux>>1;
		}		
		// Parity Odd
		wgn_parityO=0;
		wgn_data_aux=wgn_data>>(6);
		for (int i=0; i<13; i++){
			wgn_parityO^=(wgn_data_aux & 0x00000001);
			wgn_data_aux=wgn_data_aux>>1;
		}
		if (wgn_parityE==0x00 && wgn_parityO==0x01)
			// Parity Bits OK
			printf("Parity Bits OK.\n");
		else if(wgn_parityE==0x01 && wgn_parityO==0x01)
			// Parity Bit E error
			printf("Parity Bit Even Error.\n");
		else if (wgn_parityE==0x01 && wgn_parityO==0x00)
			// Parity Bit O Error
			printf("Parity Bit Odd Error.\n");
		else
			// Parity Bits Error
			printf("Parity Bits Error.\n");
		break;
	case 30:		// 30 Bit Wiegand
		// TODO: Add code to process wiegand 30b format
		break;
	case 32:		// 32 Bit Wiegand
		// TODO: Add code to process wiegand 32b format
		break;
	case 34:		// 34 Bit Wiegand
		// TODO: Add code to process wiegand 34b format
		break;
	case 35:		// 35 Bit Wiegand
		// TODO: Add code to process wiegand 35b format
		break;
	case 37:		// 37 Bit Wiegand
		// TODO: Add code to process wiegand 37b format
		break;
	case 38:		// 38 Bit Wiegand
		// TODO: Add code to process wiegand 38b format
		break;
	case 42:		// 42 Bit Wiegand
		// TODO: Add code to process wiegand 42b format
		break;
	};

	printf("\n");
}

void ProcessTamperFrame(){

	if (RxFrameBuf[3]==(unsigned char)0x00)
		printf("Rst Tamper\n");
	else if (RxFrameBuf[3]==(unsigned char)0xFF)
		printf("Set Tamper\n");
	else
		printf("Err Tamper");
}
