// OutputMode.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"

//----------------------------------------------------------
// Global Variables 
//----------------------------------------------------------

FT_STATUS ftStatus;
FT_HANDLE ftHandle;
Receiver_States RxStates;
UCHAR RxFrameBuf[256];
int idxRxFrame;

//----------------------------------------------------------
// Functions prototypes (Public)
//----------------------------------------------------------

DWORD GetDeviceList(DevList * list);
DWORD PrintDeviceList(DevList * list);
DWORD ChooseDeviceFromList(DevList * list);
void FreeAllocatedMemory(DevList * list);
void DisconnectDevice();
void Set_Tamper(FT_HANDLE device_handle);
void Reset_Tamper(FT_HANDLE device_handle);
void Send_WiegandFrame26b(FT_HANDLE device_handle);
void Send_WiegandFrame30b(FT_HANDLE device_handle);     // TODO:
void Send_WiegandFrame34b(FT_HANDLE device_handle);	    // TODO:
void Send_WiegandFrame35b(FT_HANDLE device_handle);	    // TODO:
void Send_WiegandFrame37b(FT_HANDLE device_handle);     // TODO:
void Send_WiegandFrame38b(FT_HANDLE device_handle);     // TODO:
void Send_WiegandFrame42b(FT_HANDLE device_handle);     // TODO:
void Send_WiegandFrame26b(FT_HANDLE device_handle,int facility_code,long badge_id);
void Send_ResetRequest(FT_HANDLE device_handle);
void Send_EEPROMDumpRequest(FT_HANDLE device_handle);
void PrintHelp();
void ProcessFrame();
void ProcessGPIO0Frame();
void ProcessGPIO1Frame();
void ProcessRxData(unsigned char * RxChars, unsigned int BytesRx);

//----------------------------------------------------------
// Functions prototypes (Private)
//----------------------------------------------------------

//void Send_Command(FT_HANDLE device_handle, UCHAR cmd_ID, UCHAR cmd_data_size, UCHAR * cmd_data);

//----------------------------------------------------------
// Main code
//----------------------------------------------------------
int _tmain(int argc, _TCHAR* argv[])
{
	DWORD numAvailableDevs=0;
	int rValue;
	char ch;
	UCHAR RxBuffer[256];
	DWORD RxBytes,BytesReceived;
	static DevList AvailableDevicesList;

	printf("ETConcept Engenharia de sistemas LDA\n");
	printf("W2USB Wiegand Converters Application\n");
	printf("Example 2 - Output mode\n");
	printf("\n\n List of avaiable devices: \n\n");

	numAvailableDevs=GetDeviceList(&AvailableDevicesList);

	PrintDeviceList(&AvailableDevicesList);

	rValue=ChooseDeviceFromList(&AvailableDevicesList);

	PrintHelp();

	while(rValue != -1){
		if (_kbhit()){
			ch = _getch();
			switch(ch){
			case 'T':
				printf("Set Tamper");
				Set_Tamper(ftHandle);
				break;
			case 't':
				printf("Reset Tamper");
				Reset_Tamper(ftHandle);
				break;
			case 'h':
			case 'H':
				PrintHelp();
				break;
			case 'r':
			case 'R':
				printf("Reset request ");
				Send_ResetRequest(ftHandle);
				break;
			case 'w':
			case 'W':
				printf("Send 26b Wiegand Frame \n");
				Send_WiegandFrame26b(ftHandle);
				break;
			case 'd':
			case 'D':
				printf("EEPROM Dump Request  ");
				Send_EEPROMDumpRequest(ftHandle);
				break;
			case 'x':
			case 'X':
				printf("Exit \n\n");
				rValue=-1;
				break;
			};
		}

		// Check if new data was send by the converter

        // Do your seconds based tasks here.
		FT_GetQueueStatus(ftHandle,&RxBytes); 
		if (RxBytes>0){
			ftStatus = FT_Read(ftHandle,RxBuffer,RxBytes,&BytesReceived);
			ProcessRxData(RxBuffer,BytesReceived);
			ProcessFrame();
		}
	}

	FreeAllocatedMemory(&AvailableDevicesList);

	DisconnectDevice();
	return 0;
}

DWORD GetDeviceList(DevList * list){

	FT_STATUS ftStatus;
	FT_HANDLE ftHandleTemp;
	DWORD numDevs, auxnumDevs;
	DWORD Flags;
	DWORD ID;
	DWORD Type;
	DWORD LocId;
	UCHAR SerialNumber[16];
	UCHAR Description[64];							

	// Create the device information list
	ftStatus = FT_CreateDeviceInfoList(&numDevs);
	auxnumDevs=0;
	while (auxnumDevs < numDevs && auxnumDevs<MAXIMUM_NUM_DEVICES ) { 
		ftStatus = FT_GetDeviceInfoDetail(auxnumDevs, &Flags, &Type, &ID, &LocId, SerialNumber, Description, &ftHandleTemp);
		if (ftStatus == FT_OK) { 
			list->SN_List[auxnumDevs]=new UCHAR[sizeof(SerialNumber)+1];
			strcpy_s((char *)list->SN_List[auxnumDevs],sizeof(SerialNumber),(char *)SerialNumber);
		}
		auxnumDevs++;
	}
	list->numDevs=numDevs;
	return numDevs;
}

DWORD PrintDeviceList(DevList * list){
	DWORD numDevs=0;

	while (numDevs < list->numDevs && list->SN_List[numDevs]!=NULL ) { 
		printf("%d) SN=%s \n",numDevs+1,list->SN_List[numDevs]);
		numDevs++;
	}
	return numDevs;		// Return thenumber of devices printed
}

DWORD ChooseDeviceFromList(DevList * list){

	int devIdx=-1;

	while(devIdx<1 || devIdx>(signed int)list->numDevs){
		printf("\n Choose from (1-%d) :",list->numDevs);
		scanf_s("%d",&devIdx);
	}

	// Connect to the Device using the SerialNumber
	ftStatus=FT_OpenEx(list->SN_List[devIdx-1],FT_OPEN_BY_SERIAL_NUMBER,&ftHandle);
	if (ftStatus != FT_OK){
		printf("Erro! Error code %d (USB Device %s isn't plug to USB port)\n\r",ftStatus,list->SN_List[devIdx-1]);
		return NULL;
	}

	// *************************************
	// Set the Device communication settings
	// *************************************

	//	Set BaudRate to 9600  (Baudrate should be maintain at 9600!)
	ftStatus = FT_SetBaudRate(ftHandle, 9600); // Set baud rate to 9600
	if (ftStatus != FT_OK){
		printf("Erro! Error code %d (Couldn't set the BaudRate)\n\r",ftStatus);
		return ftStatus;
	}
	//	Set Databits to 8bits, StopBits to 1bit and Parity to None
	ftStatus = FT_SetDataCharacteristics(ftHandle, FT_BITS_8, FT_STOP_BITS_1, FT_PARITY_NONE);
	if (ftStatus != FT_OK){
		printf("Erro! Error code %d (Couldn't set Data bits or Stop bits or Parity bit)\n\r",ftStatus);
		return ftStatus;
	}
	printf("\n Connected to device SN=%s\n",list->SN_List[devIdx-1]);
	return devIdx;
}

void FreeAllocatedMemory(DevList * list){
	for (int nums=list->numDevs; nums>0; nums--)
		delete [] list->SN_List[nums];
}


void DisconnectDevice(){

	if(ftHandle)							// If connected ato any device
		ftStatus=FT_Close(ftHandle);		// disconect
	if (ftStatus != FT_OK){
		
	}
}

void Send_Command(FT_HANDLE device_handle, UCHAR cmd_ID, UCHAR cmd_data_size, UCHAR * cmd_data){
	char TmpRxBuffer[21];
	DWORD BytesWritten;
	TmpRxBuffer[0]=0x55;
	TmpRxBuffer[1]=0x55;
	TmpRxBuffer[2]=cmd_ID;
	memcpy(TmpRxBuffer+3,(char *)cmd_data,cmd_data_size);
	TmpRxBuffer[cmd_data_size+3]=0x00;
	TmpRxBuffer[cmd_data_size+4]=0x0D;

	ftStatus = FT_Write(device_handle, TmpRxBuffer, cmd_data_size+5, &BytesWritten); 
	if (ftStatus == FT_OK) { // FT_Write OK 
		printf("  Ok\n");
	} else { 
	// FT_Write Failed 
		printf("  Nok\n");
	}
}

void Set_Tamper(FT_HANDLE device_handle){
	UCHAR cmd_data[2];

	cmd_data[0]=0xFF;
	cmd_data[1]=0x00;

	Send_Command(device_handle, 0x02, 1, cmd_data);
}

void Reset_Tamper(FT_HANDLE device_handle){
	UCHAR cmd_data[2];

	cmd_data[0]=0x00;
	cmd_data[1]=0x00;

	Send_Command(device_handle, 0x02, 1, cmd_data);
}

void Send_ResetRequest(FT_HANDLE device_handle){
	UCHAR cmd_data[2];

	cmd_data[0]=0x00;
	cmd_data[1]=0x00;

	Send_Command(device_handle, 0x0A, 1, cmd_data);
}

void Send_EEPROMDumpRequest(FT_HANDLE device_handle){
	UCHAR cmd_data[2];
	unsigned char RxBuffer[256];
	DWORD RxBytes,BytesReceived;

	// Clear Rx buffer
	FT_SetTimeouts(device_handle,1000,0);		// Set Read timeout to 1 second
	FT_GetQueueStatus(device_handle,&RxBytes); 
	if (RxBytes>0)
		ftStatus = FT_Read(device_handle,RxBuffer,RxBytes,&BytesReceived); 


	cmd_data[0]=0x00;
	cmd_data[1]=0x00;

	Send_Command(device_handle, 0x0B, 1, cmd_data);

	::Sleep(100);
	// Wait for the response
	FT_SetTimeouts(device_handle,1000,0);		// Set Read timeout to 1 second
	FT_GetQueueStatus(device_handle,&RxBytes); 
	if (RxBytes>0){
		ftStatus = FT_Read(ftHandle,RxBuffer,RxBytes,&BytesReceived); 
		if (ftStatus == FT_OK) { // FT_Read OK 
			// TODO: Do not assumes that received data is ok
			printf("\n");
			for (unsigned int i=0; i < RxBytes ; i++){
				printf("%02X ",(unsigned char)RxBuffer[i]);
				if ( ((i+1) % 13)== 0 ){
					printf("\n");
				}
			}
			printf("\n");
		} else { 
			// FT_Read Failed 
			printf("Error reading EEPROM \n");
		} 
	}

}

void PrintHelp(){
	printf("\n\n");
	printf ("Press 'T' to set GPIO0 to high level\n");
	printf ("Press 't' to set GPIO0 to low level\n");
	printf ("Press 'r' or 'R' to send a Reset device\n");
	printf ("Press 'd' or 'D' to read the device EEPROM\n");
	printf ("Press 'x' or 'X' to exit\n");
	printf("\n\n");
}

void ProcessRxData(unsigned char * RxChars, unsigned int BytesRx){
	unsigned int i;	

	i=0;
	while (i<BytesRx){
		switch(RxStates){
			case WAIT_SYNC:
				// Verify if any of the received Chars are 0x55
				while(RxChars[i]!=(unsigned char)0x55 && i<BytesRx) i++;
				// If exist a 0x55 char Check if char next to it is also a 0x55
				if (RxChars[i]==(unsigned char)0x55 && BytesRx>i){
					// First Sync char detected
					// Copy to the RxFrameBuf
					RxFrameBuf[idxRxFrame++]=RxChars[i];
					// Check if the adjacent char is also Sync char
					if((i+1)<BytesRx){	// Sync char isn't the last of this block
						// Check if the adjacent char is also Sync char
						// If true, Preliminary Start of the frame detected change to WAIT_END state
						i++;
						if (RxChars[i]==(unsigned char)0x55){
							RxFrameBuf[idxRxFrame++]=RxChars[i++];
							RxStates=WAIT_END;
						}
					}else{
						// First Sync char detected in the end of this block wait for the second Sync
						RxStates=WAIT_SYNC2;
					}
				}else{
					// else maintain the current state
				}
				break;
			case WAIT_SYNC2:
				if (RxChars[i]==(unsigned char)0x55){
					RxFrameBuf[idxRxFrame++]=RxChars[i++];
					RxStates=WAIT_END;
				}else{
					idxRxFrame=0;			// False Alarm clear RxFrameBuf
					RxStates=WAIT_SYNC;
				}
				i++;
				break;
			case WAIT_END:
				// Check if any of the received chars is 0x0D
				while(RxChars[i]!=(unsigned char)0x0D && i<BytesRx && idxRxFrame<20){
					RxFrameBuf[idxRxFrame++]=RxChars[i++];	// Store Char
				}
				if (idxRxFrame>=20){
					// End of frame wasnt detected
					idxRxFrame=0;			// False Alarm clear RxFrameBuf
					// Return to Wait sync
					RxStates=WAIT_SYNC;
				}else if (i>=BytesRx){		// End of the block
					// Maintain Rx state
				}else{
					// End of frame detected
					RxFrameBuf[idxRxFrame++]=RxChars[i++];
					// check frame coerence Dimension x distance between 
					if ( ( RxFrameBuf[2] == (unsigned char)0x01 && idxRxFrame==20) ||
						 ( RxFrameBuf[2] == (unsigned char)0x02 && idxRxFrame==6) ||
						 ( RxFrameBuf[2] == (unsigned char)0x03 && idxRxFrame==6) ||
						 ( RxFrameBuf[2] == (unsigned char)0x04 && idxRxFrame==6) ||
						 ( RxFrameBuf[2] == (unsigned char)0x09 && idxRxFrame==7) ||
						 ( RxFrameBuf[2] == (unsigned char)0x0A && idxRxFrame==6) ||
						 ( RxFrameBuf[2] == (unsigned char)0x0B && idxRxFrame==6)
					   ){
						// Sync chars and end frame char and dim ok
						// Change to PROCESSING_DATA
						RxStates=PROCESSING_DATA;
					}else{
						// Frame isnt over. Maintain same state

					}
				}
			break;
			case PROCESSING_DATA:
				ProcessFrame();
			break;
		};
	}
}

void ProcessFrame(){
	if (RxStates==PROCESSING_DATA){
		switch(RxFrameBuf[2]){
		// (Only for Input mode)
		//case 0x01:	// Wiegand Frame Received
		//	ProcessWiegandFrame();
		//	break;
		//case 0x02:  // Change notification of the Tamper
		//	ProcessTamperFrame();
		//	break;

		case 0x03:  // Change notification of the GPIO0
			ProcessGPIO0Frame();		
			break;
		case 0x04:  // Change notification of the GPIO1
			ProcessGPIO1Frame();
			break;
		case 0x09:	// Change EEPROM data
			// do nothing
			break;

			default:
				break;
		};
		idxRxFrame=0;
		RxStates=WAIT_SYNC;
	}
}

void ProcessGPIO0Frame(){

	if (RxFrameBuf[3]==(unsigned char)0x00)
		printf("Rst GPIO0\n");
	else if (RxFrameBuf[3]==(unsigned char)0xFF)
		printf("Set GPIO0\n");
	else
		printf("Err GPIO0");
}

void ProcessGPIO1Frame(){

	if (RxFrameBuf[3]==(unsigned char)0x00)
		printf("Rst GPIO1\n");
	else if (RxFrameBuf[3]==(unsigned char)0xFF)
		printf("Set GPIO1\n");
	else
		printf("Err GPIO1");
}

void Send_WiegandFrame26b(FT_HANDLE device_handle){
	int facility_code=-1;
	unsigned char parity_mask;
	long badge_id=-1;
	UCHAR parity_E,parity_O;
	UCHAR cmd_data[16];

	// Clear Send Buffer
	memset(cmd_data,0x00,16);

	while(facility_code<0 || facility_code>255){
		printf("Facility code [0-255] :");
		scanf_s("%d",&facility_code);
	}
	printf("\n");
	while(badge_id<0 || badge_id>65535){
		printf("Badge_ID [0-65535] :");
		scanf_s("%d",&badge_id);
	}
	printf("\n");

	cmd_data[0]=26;												// 26b Wiegand Format
	cmd_data[1]=(UCHAR)(facility_code>>1);						// Add 7 MSB of the facility code without parity
	cmd_data[2]=(UCHAR)((facility_code<<7) | (badge_id >> 9));
	cmd_data[3]=(UCHAR)((badge_id >> 1));
	cmd_data[4]=(UCHAR)((badge_id << 7));
	// Calc parity bits
	parity_E=0;
	for (int idx_bit=0;idx_bit<13;idx_bit++){
		int idx_byte=idx_bit / 8;
		parity_mask=(0x80 >> (idx_bit-idx_byte*8));
		if (cmd_data[idx_byte+1] & parity_mask)
			parity_E++;
	}
	parity_O=0;
	for (int idx_bit=13;idx_bit<26;idx_bit++){
		int idx_byte=idx_bit / 8;
		parity_mask=(0x80 >> (idx_bit-idx_byte*8));
		if (cmd_data[idx_byte+1] & parity_mask)
			parity_O++;
	}
	// Add parity bits
	cmd_data[1]|=(parity_E<<7);
	cmd_data[4]|=((~parity_O & 0x01)<<6);
	Send_Command(device_handle, 0x01, 15, cmd_data);
	printf("Wiegand Frame sent \n");
}

void Send_WiegandFrame26b(FT_HANDLE device_handle,int facility_code,long badge_id){
	unsigned char parity_mask;
	UCHAR parity_E,parity_O;
	UCHAR cmd_data[16];

	// Clear Send Buffer
	memset(cmd_data,0x00,16);

	cmd_data[0]=26;												// 26b Wiegand Format
	cmd_data[1]=(UCHAR)(facility_code>>1);						// Add 7 MSB of the facility code without parity
	cmd_data[2]=(UCHAR)((facility_code<<7) | (badge_id >> 9));
	cmd_data[3]=(UCHAR)((badge_id >> 1));
	cmd_data[4]=(UCHAR)((badge_id << 7));
	// Calc parity bits
	parity_E=0;
	for (int idx_bit=0;idx_bit<13;idx_bit++){
		int idx_byte=idx_bit / 8;
		parity_mask=(0x80 >> (idx_bit-idx_byte*8));
		if (cmd_data[idx_byte+1] & parity_mask)
			parity_E++;
	}
	parity_O=0;
	for (int idx_bit=13;idx_bit<26;idx_bit++){
		int idx_byte=idx_bit / 8;
		parity_mask=(0x80 >> (idx_bit-idx_byte*8));
		if (cmd_data[idx_byte+1] & parity_mask)
			parity_O++;
	}
	// Add parity bits
	cmd_data[1]|=(parity_E<<7);
	cmd_data[4]|=((~parity_O & 0x01)<<6);
	Send_Command(device_handle, 0x01, 15, cmd_data);
	printf("Wiegand Frame sent \n");
}

void Send_WiegandFrame30b(FT_HANDLE device_handle){
// TODO:
}

void Send_WiegandFrame34b(FT_HANDLE device_handle){
// TODO:
}

void Send_WiegandFrame35b(FT_HANDLE device_handle){
// TODO:
}

void Send_WiegandFrame37b(FT_HANDLE device_handle){
// TODO:
}

void Send_WiegandFrame38b(FT_HANDLE device_handle){
// TODO:
}

void Send_WiegandFrame42b(FT_HANDLE device_handle){
// TODO:
}
